#! -*- coding: utf_8 -*-

import enum, random, os, re, json, inspect, asyncio
from io import StringIO
import xml.etree.ElementTree as ET
from xml.dom import minidom
from html.parser import HTMLParser

from . import __version__ as oxi_version
from .utils import SmartDict, aopen, dual_mode
from .server import OxiProtocol

oxi_version
       
#############################################################################################################################

class Carets(enum.Enum):
    CARET_UP = "&#9650;"
    CARET_DOWN = "&#9660;"

MenuColors = {
    'MAINBACKNORMAL':  "#676767",
    'MAINBACKSELECTED':  "steelblue",
    'MAINFORENORMAL':  "#ffffff",
    'MAINFORESELECTED':  "#020582",
    'ITEMBACKNORMAL':  "#f9f9f9",
    'ITEMBACKSELECTED':  "#dddddd",
    'ITEMFORENORMAL':  "#000000",
    'ITEMFORESELECTED':  "#020281",
}


class Menu:
    """
    Handles menu display stuff
    """

    mainBack = MenuColors.get("MAINBACKNORMAL")
    mainBackSelected = MenuColors.get("MAINBACKSELECTED")
    itemBack = MenuColors.get("ITEMBACKNORMAL")
    itemBackSelected = MenuColors.get("ITEMBACKSELECTED")

    style = """
    .navbar {{
        overflow: hidden;
        background-color: {0};
    }}

    .navheader {{
        color: #e4e4e4; 
        font-size: 150%; 
        font-weight: bold;
        float: left;
        display: flex;
        flex-direction: row;
        align-items: center;
        padding: 10px;
    }}

    .navbar a {{
        float: left;
        font-size: 16px;
        color: white;
        text-align: center;
        padding: 14px 16px;
        text-decoration: none;
    }}

    .dropdown {{
        float: left;
        overflow: hidden;
    }}

    .dropdown .dropbtn {{
        font-size: 16px;
        border: none;
        outline: none;
        color: white;
        padding: 14px 16px;
        background-color: inherit;
        font-family: inherit;
        margin: 0;
    }}

    .navbar a:hover, .dropdown:hover .dropbtn {{
        background-color: {1};
    }}

    .dropdown-content {{
        display: none;
        position: absolute;
        background-color: {2};
        min-width: 160px;
        box-shadow: 0px 8px 16px 0px rgba(0,0,0,0.2);
        z-index: 1;
    }}

    .dropdown-content a {{
        float: none;
        color: black;
        padding: 12px 16px;
        text-decoration: none;
        display: block;
        text-align: left;
    }}

    .dropdown-content a:hover {{
        background-color: {3};
    }}

    .dropdown:hover .dropdown-content {{
        display: block;
    }}
    """

    @classmethod
    def getStyle(cls, mB = mainBack, mBS= mainBackSelected, iB= itemBack, iBS= itemBackSelected):
        return cls.style.format(mB, mBS, iB, iBS)


class MenuItem(Menu):
    def __init__(self, label: str, href: str):
        self.label = label
        self.href = href

    def __repr__(self):
        return '<a href="{0}">{1}</a>'.format(self.href, self.label)

class MenuSep(MenuItem):
    def __init__(self):
        super().__init__('', '')

    def __repr__(self):
        return '<hr>'


class DropdownMenu(Menu):
    def __init__(self, label="Dropdown", items: list[MenuItem] = None):
        self.label = label
        self.items: list[MenuItem] = items or []

    def addItem(self, item: MenuItem):
        if not isinstance(item, (MenuItem, DropdownMenu)):
            return False
        else:
            self.items.append(item)
            return True

    def __repr__(self):
        parts = []
        prefix = f"""
        <div class="dropdown">
            <button class="dropbtn">{self.label}
            <!--<i class="fa fa-caret-down"></i>-->
            <i style="font-size: 75%;">&nbsp;&#9660;</i>
            </button>
            <div class="dropdown-content">
        """
        parts.append(prefix)
        for i in self.items:
            parts.append(str(i))
        suffix = """
            </div>
        </div>
        """
        parts.append(suffix)
        return "".join(parts)


class MenuBar:

    def __init__(self, items: list[Menu] = None, mB=Menu.mainBack, mBS=Menu.mainBackSelected, iB=Menu.itemBack, iBS=Menu.itemBackSelected):
        self.items = items or []
        self.mB = mB
        self.mBS = mBS
        self.iB = iB
        self.iBS = iBS

    def setBackground(self, color):
        self.mB = color
        
    def addItem(self, item: MenuItem):
        if not isinstance(item, (MenuItem, DropdownMenu)):
            return False
        else:
            self.items.append(item)
            return True

    def __repr__(self):
        parts = []
        for i in self.items:
            parts.append(str(i))
        joined = "".join(parts)
        return f'<style>{Menu.getStyle(self.mB, self.mBS, self.iB, self.iBS)}</style><div class="navbar">{joined}</div>'


class NavBar(MenuBar):

    def __init__(self, head: str = None, items: list[Menu] = None):
        super().__init__(items)
        self.head = head or ''

    def __repr__(self):
        parentrepr = super().__repr__()
        replace = f'<div class="navbar"><div class="navheader">&nbsp;{self.head}&nbsp;&nbsp;&nbsp;</div>'
        return parentrepr.replace('<div class="navbar">', replace)
    

class Head:
    def __init__(self, title: str, links: list[str] = None, styles: list[str] = None, scripts: list[str] = None):
        self.title = title
        self.links = links or []
        self.styles = styles or []
        self.scripts = scripts or []

    @property
    def links_list(self) -> str:
        return('\n'.join(self.links))

    @property
    def styles_list(self) -> str:
        return('\n'.join(self.styles))

    @property
    def scripts_list(self) -> str:
        return('\n'.join(self.scripts))

    def add_item(self, item_type: str = 'script', content: str = ''):
        if item_type == 'link':
            target_list = self.links
        elif item_type == 'style':
            target_list = self.styles
        elif item_type == 'script':
            target_list = self.scripts
        else:
            return False
        target_list.append(f"{content}")
        return True
    
    def __repr__(self):
        return """
        <head>
            <meta charset="utf-8">
            <title>{0}</title>
            <meta name="viewport" content="width=device-width, initial-scale=1.0, shrink-to-fit=no">
            <!--LINKS-->\n{1}
            <!--STYLES-->\n{2}
            <!--SCRIPTS-->\n{3}
        </head>
        """.format(self.title, self.links_list, self.styles_list, self.scripts_list )


class Body:
    def __init__(self, navbar: NavBar = None, font_family: str = "sans-serif"):
        self.navbar = navbar
        self.font_family = font_family

    def __repr__(self):
        retval =  '<body style="font-family: ' + self.font_family + ';">'
        retval += '<div id="main">'
        retval += str(self.navbar) if self.navbar else ''
        retval += ' <div id="contents">'
        retval += '  {{ contents }}'
        retval += '</div>'
        retval += '</div>'
        retval += '</body>'
        
        return retval


class HTMLApp:
    def __init__(self, head: Head, body: Body, lang: str = 'en'):
        self.lang = lang
        self.head = head
        self.body = body

    def __repr__(self):
        return f"""
        <!DOCTYPE html>
        <html lang={self.lang}>
        {repr(self.head)}
        {repr(self.body)}
        </html>
        """


#############################################################################################################################

######################################################################################

# Templates related code

class CodeBuilder:
    """Build source code conveniently."""

    def __init__(self, indent=0):
        self.code = []
        self.indent_level = indent

    def add_line(self, line):
        """Add a line of source to the code.
        Indentation and newline will be added for you, don't provide them.
        """
        self.code.extend([" " * self.indent_level, line, "\n"])

    INDENT_STEP = 4      # PEP8 says so!

    def indent(self):
        """Increase the current indent for following lines."""
        self.indent_level += self.INDENT_STEP

    def dedent(self):
        """Decrease the current indent for following lines."""
        self.indent_level -= self.INDENT_STEP

    def add_section(self):
        """Add a section, a sub-CodeBuilder."""
        section = CodeBuilder(self.indent_level)
        self.code.append(section)
        return section

    def __str__(self):
        return "".join(str(c) for c in self.code)

    def get_globals(self):
        """Execute the code, and return a dict of globals it defines."""
        # A check that the caller really finished all the blocks they started.
        assert self.indent_level == 0
        # Get the Python source as a single string.
        python_source = str(self)
        # Execute the source, defining globals, and return them.
        global_namespace = {}
        exec(python_source, global_namespace)
        return global_namespace


class TemplateSyntaxError(BaseException):
    pass


class TemplateLight:

    test_tpl = """
        <h2> Hello, I am {{ user  }}. </h2>
        <p>These are my favourite teams, in no particular order.<p>
        <p>
          <ul>
            {% for team in teams %}
              <li> {{ team }} </li>
            {% endfor %}
          </ul>
        </p>
        """

    def __init__(self, text, **contexts):
        """Construct a TemplateLight with the given `text`.
        `contexts` are key-value pairs to use for future renderings.
        These are good for filters and global values.
        """
        self._template_text = text
        self.context = {}
        # for context in contexts:
        # self.context.update(context)
        self.context.update(contexts)
        self.all_vars = set()
        self.loop_vars = set()

        code = CodeBuilder()
        code.add_line("def render_function(context, do_dots):")
        code.indent()
        vars_code = code.add_section()
        code.add_line("result = []")
        code.add_line("append_result = result.append")
        code.add_line("extend_result = result.extend")
        code.add_line("to_str = str")

        buffered = []

        def flush_output():
            """Force `buffered` to the code builder."""
            if len(buffered) == 1:
                code.add_line("append_result(%s)" % buffered[0])
            elif len(buffered) > 1:
                code.add_line("extend_result([%s])" % ", ".join(buffered))
            del buffered[:]

        ops_stack = []
        text = text.replace(",", " , ")
        tokens = re.split(r"(?s)({{.*?}}|{%.*?%}|{#.*?#})", text)

        for token in tokens:
            if token.startswith('{#'):
                # Comment: ignore it and move on.
                continue
            elif token.startswith('{{'):
                # An expression to evaluate.
                expr = self._expr_code(token[2:-2].strip())
                buffered.append("to_str({0})".format(expr))
            elif token.startswith('{%'):
                # Action tag: split into words and parse further.
                flush_output()
                words = token[2:-2].strip().split()
                if words[0] == 'if':
                    # An if statement: evaluate the expression to determine if.
                    # if len(words) != 2:
                    #    self._syntax_error("Don't understand if", token)
                    ops_stack.append('if')
                    code.add_line("if {0}:".format(
                        self._expr_code(' '.join(words[1:]))))
                    code.indent()
                elif words[0] == 'elif':
                    # An elif statement: evaluate the expression to determine else.
                    #print("Uso de 'else' en el template detectado.")
                    # if len(words) != 2:
                    #    self._syntax_error("Don't understand elif", token)
                    if not ops_stack:
                        self._syntax_error(
                            "'elif' without previous 'if'", token)
                    start_what = ops_stack.pop()
                    if (start_what != "if"):
                        self._syntax_error(
                            "'elif' without previous 'if'", token)
                    ops_stack.append('if')
                    code.dedent()
                    code.add_line("elif {0}:".format(
                        self._expr_code(' '.join(words[1:]))))
                    code.indent()
                elif words[0] == 'else':
                    # An else statement: evaluate the expression to determine else.
                    #print("Uso de 'else' en el template detectado.")
                    if len(words) != 1:
                        self._syntax_error("Don't understand else", token)
                    if not ops_stack:
                        self._syntax_error(
                            "'Else' without previous 'if'", token)
                    start_what = ops_stack.pop()
                    if (start_what != "if"):
                        self._syntax_error(
                            "'Else' without previous 'if'", token)
                    ops_stack.append('else')
                    code.dedent()
                    code.add_line("else:")
                    code.indent()
                elif words[0] == 'for':
                    # A loop: iterate over expression result.
                    if words[-2] != 'in':
                        self._syntax_error("Don't understand for", token)
                    ops_stack.append('for')
                    loopvars = list(filter(lambda x: x != ',', words[1:-2]))
                    for loopvar in loopvars:
                        self._variable(loopvar, self.loop_vars)
                    deco_loopvars = list(map(lambda v: f"c_{v}", loopvars))
                    line_to_add = "for {0} in {1}:".format(
                        " , ".join(deco_loopvars), self._expr_code(words[-1]))
                    code.add_line(line_to_add)
                    code.indent()
                elif words[0].startswith('end'):
                    # Endsomething.  Pop the ops stack.
                    if len(words) != 1:
                        self._syntax_error("Don't understand end", token)
                    end_what = words[0][3:]
                    if not ops_stack:
                        self._syntax_error("Too many ends", token)
                    start_what = ops_stack.pop()
                    if (start_what != end_what) and (start_what != "else" or end_what != "if"):
                        self._syntax_error("Mismatched end tag", end_what)
                    code.dedent()
                else:
                    self._syntax_error("Don't understand tag", words[0])
            else:
                # Literal content.  If it isn't empty, output it.
                if token:
                    buffered.append(repr(token))

        if ops_stack:
            self._syntax_error("Unmatched action tag", ops_stack[-1])

        flush_output()

        for var_name in self.all_vars - self.loop_vars:
            vars_code.add_line("c_%s = context[%r]" % (var_name, var_name))

        code.add_line("return ''.join(result)")
        code.dedent()

        self._code = code
        self._render_function = code.get_globals()['render_function']

    @staticmethod
    def _is_string(name):
        pattrn = r"""^(\"|\')(.*?)\1$"""
        return re.match(pattrn, name)

    @staticmethod
    def _is_reserved(name):
        if name == "true":
            name = "True"
        if name == "false":
            name = "False"
#        if name == "null" or name == "nil":
#            name = "None"
        return name in ["|", "if", "else", "and", "or", "not", "in", "is", "True", "False", "None"]

    @staticmethod
    def _is_variable(name):
        if TemplateLight._is_reserved(name) or TemplateLight._is_string(name):
            return False
        pattrn = r"(?P<varname>[_a-zA-Z][_a-zA-Z0-9]*)(?P<subscript>\[(?P<subvar>.+)\])?$"
        return re.match(pattrn, name)

    def _variable(self, name, vars_set):
        """Track that `name` is used as a variable.
        Adds the name to `vars_set`, a set of variable names.
        Raises an syntax error if `name` is not a valid name.
        """
        m = self._is_variable(name)
        if not m:
            self._syntax_error("Not a valid name", name)
        d = m.groupdict()
        vars_set.add(d.get("varname"))
        #subvar = d.get("subvar")
        # if subvar and self._is_variable(subvar):
        #    vars_set.add(self._expr_code(subvar))

    def _expr_code(self, expr):
        """Generate a Python expression for `expr`."""
        nonisobar = r"[^\s]\|[^\s]"
        if re.findall(nonisobar, expr):
            pipes = expr.split("|")
            code = self._expr_code(pipes[0])
            for func in pipes[1:]:
                self._variable(func, self.all_vars)
                code = "c_%s(%s)" % (func, code)
        elif "." in expr:
            dots = expr.split(".")
            code = self._expr_code(dots[0])
            args = ", ".join(repr(d) for d in dots[1:])
            code = "do_dots(%s, %s)" % (code, args)
        else:
            subexprs = expr.split()
            code = ""
            for subexpr in subexprs:
                m = self._is_variable(subexpr)
                if m:
                    self._variable(subexpr, self.all_vars)
                    d = m.groupdict()
                    varname = f"c_{d.get('varname')}"
                    subscript = d.get('subscript', '')
                    if subscript:
                        subvar = d.get('subvar')
                        if subvar and self._is_variable(subvar):
                            subscript = subscript.replace(
                                subvar, f"c_{subvar}")
                    code += "{0}{1} ".format(varname,
                                             subscript if subscript else "")
                else:
                    code += f"{subexpr} "
        return code

    def _syntax_error(self, msg, thing):
        """Raise a syntax error using `msg`, and showing `thing`."""
        raise TemplateSyntaxError("%s: %r" % (msg, thing))

    def _do_dots(self, value, *dots):
        """Evaluate dotted expressions at runtime."""
        for dot in dots:
            try:
                value = getattr(value, dot)
            except AttributeError:
                if isinstance(value, (list, tuple)):
                    if isinstance(dot, str) and dot.isnumeric():
                        dot = int(dot)
                value = value[dot]
            if callable(value):
                value = value()
        return value

    def render(self, **context):
        """Render this template by applying it to `context`.
        `context` is a dictionary of values to use in this rendering.
        """
        # Make the complete context we'll use.
        render_context = dict(self.context)
        if context:
            render_context.update(context)
        return self._render_function(render_context, self._do_dots)
    
    template_filters = {}

TemplateLight.template_filters['title'] = str.title
TemplateLight.template_filters['capitalize'] = str.capitalize
TemplateLight.template_filters['upper'] = str.upper
TemplateLight.template_filters['lower'] = str.lower 
TemplateLight.template_filters['toJson'] = lambda x: json.dumps(x, ensure_ascii=False)

def get_template_dir():
    templates_root = OxiProtocol.config.get(
        'template-dir', 'templates')
    return os.path.join(os.getcwd(), templates_root)

def get_template_fullpath(template_file):
    return os.path.join(get_template_dir(), template_file)

def _load_template_sync(template_filename):
    path = get_template_fullpath(template_filename)
    with open(path, "r", encoding="utf-8") as fp:
        return fp.read()

def load_template(template_filename):
    path = get_template_fullpath(template_filename)
    with open(path, "r", encoding="utf-8") as fp:
        return fp.read()

async def load_template_async(template_filename):
    return await asyncio.to_thread(load_template, template_filename)
    
def preprocess_template(tpl_str=TemplateLight.test_tpl):
    ftpl = StringIO(tpl_str)
    lines = ftpl.readlines()
    ftpl.close()
    for index, line in enumerate(lines):
        stripline = line.strip()
        # m = re.match(r"^\{\#include\s+(?P<inc_file>.+?)\s+\#\}$", stripline)
        if stripline.startswith('{#include') and stripline.endswith("#}"):
            # m = re.match(r"^(.+?)\s+(.+?)\s+(\#\})$", stripline)
            m = re.split(r"\s+", stripline)
            if m:
                # inc_file = m.groupdict().get("inc_file")
                # inc_file = m.groups()[1]
                inc_file = m[1]
                if not inc_file:
                    raise TemplateSyntaxError(
                        "Include directiva must refer to a file")
                fullpath = get_template_fullpath(
                    inc_file.replace("\"", "").replace("'", ""))
                if os.path.exists(fullpath) and os.path.isfile(fullpath):
                    fp = open(fullpath)
                    new_tpl_str = fp.read()
                    fp.close()
                    replace_line = preprocess_template(
                        new_tpl_str)
                    lines[index] = replace_line
            else:
                raise TemplateSyntaxError(
                    "Included file {0} in line {1} does not exist.".format(inc_file, index))
    return "".join(lines)

def compile_template(tpl_str=TemplateLight.test_tpl):
    if not tpl_str:
        return None
    words = tpl_str.split()
    if len(words) == 1:
        fullpath = get_template_fullpath(tpl_str)
        if os.path.exists(fullpath) and os.path.isfile(fullpath):
            # fp = open(fullpath)
            # tpl_str = fp.read()
            # fp.close()
            tpl_str = load_template(tpl_str)

    return TemplateLight(preprocess_template(tpl_str),
                            **TemplateLight.template_filters)

def render_template(tpl_str=TemplateLight.test_tpl, **kw):
    if tpl_str.__class__.__name__ == "TemplateLight":
        return tpl_str.render(**kw)
    elif tpl_str.__class__.__name__ == "str":
        compiled = compile_template(preprocess_template(tpl_str))
        if compiled:
            return compiled.render(**kw)
        else:
            return None
    else:
        return None

    # End of template related stuff

# End of templates related code

#############################################################################################################################

#############################################################################################################################

# Special templates

swiss_army_knife_template = """
{{% if full %}}
    <{{{{parentElement}}}} class="{{{{parentClass}}}}" {{% if size %}} size="{{{{ size }}}}" {{% endif%}}>
{{% endif %}}
{header}
{{% for record in records %}}
    {{% if childElement == 'option' %}}
        <{{{{childElement}}}} {{% if record['{idColumn}'] == {selected} %}} selected {{% endif %}} value="{{{{ record['{idColumn}'] }}}}" data-id="{{{{ record['{idColumn}'] }}}}" data-json='{{{{ record|toJson }}}}'>
            {{{{ record['{nameColumn}'] }}}}
        </{{{{childElement}}}}>
    {{% elif childElement == 'li' %}}
        <{{{{childElement}}}} data-id="{{{{ record['{idColumn}'] }}}}" data-json='{{{{ record|toJson }}}}'>
            <span {{% if record['{idColumn}'] == {selected} %}} class="selected" {{% endif %}}>{{{{ record['{nameColumn}'] }}}}</span>
        </{{{{childElement}}}}>
    {{% elif childElement == 'tr' %}}
        <{{{{childElement}}}} data-id="{{{{ record['{idColumn}'] }}}}" data-json='{{{{ record|toJson }}}}' {{% if record['{idColumn}'] == {selected} %}} class="selected" {{% endif %}}>
            {{%for key in keyz %}}
                <{{{{grandchildElement}}}}> 
                    {{{{ record[key] }}}}
                </{{{{grandchildElement}}}}>
            {{%endfor%}}
        </{{{{childElement}}}}>
    {{% elif childElement == 'p' %}}
        <{{{{childElement}}}} data-id="{{{{ record['{idColumn}'] }}}}" data-json='{{{{ record|toJson }}}}' {{% if record['{idColumn}'] == {selected} %}} class="selected" {{% endif %}}>
            {{%for key in keyz %}}
                <{{{{grandchildElement}}}}> 
                    {{{{ record[key] }}}}
                </{{{{grandchildElement}}}}>
            {{%endfor%}}
        </{{{{childElement}}}}>
    {{% else %}}
        <p data-id="{{{{ record['{idColumn}'] }}}}" data-json='{{{{ record|toJson }}}}'>{{{{record['{nameColumn}']}}}}</p>
    {{% endif %}}
{{% endfor %}}
{footer}

{{% if full %}}
    </{{{{parentElement}}}}>
{{% endif %}}

"""

def swiss_army_knife(records, **kwargs):
    options = SmartDict({
        'parentElement': 'select',
        'childElement': 'option',
        'grandchildElement': 'span',
        'header': '',
        'footer': '',
        'selected': 0,
        'idColumn': 'id',
        'nameColumn': 'name',
        'titleColumn': 'description',
        'full': True,
        'parentClass': '',
        'emptyMsg': 'No records found.',
        'selected': 0,
        'size': None
    })
    options.update(kwargs)
    if not len(records):
        return f"""
        <p class='{options.parentClass}'>{options.emptyMsg}</p>
        """
    options.keyz = records[0].keys()
    options.headers = [key.title() for key in options.keyz]
    
    if options.parentElement.lower() == 'table':
        options.childElement = 'tr'
        options.grandchildElement = 'td'
        options.parentClass = options.parentClass or 'zebra'
        options.header = f"""
        <thead>
        <tr>
        {'<th>' + '</th><th>'.join(options.headers) + '</th>'}
        </tr>
        </thead>
        <tbody>
        """
        options.footer = '</tbody>'
    elif options.parentElement.lower() == 'div':
        options.childElement = 'p'
        options.grandchildElement = 'span'
        options.parentClass = options.parentClass or 'panel'
    elif options.parentElement.lower() == 'ul':
        options.childElement = 'li'
        options.grandchildElement = 'a'
        options.parentClass = options.parentClass or 'big'
    elif options.parentElement.lower() == 'select':
        options.childElement = 'option'
        options.grandchildElement = 'span'
        options.parentClass = options.parentClass or 'big'

    if options.nameColumn not in options.keyz:
        listkeyz = list(options.keyz)
        if len(listkeyz) > 1:
            options.nameColumn = list(options.keyz)[1] 
        else:
            options.nameColumn = list(options.keyz)[0]

    formatted_template = swiss_army_knife_template.format(**options)
    rendered = render_template(formatted_template, records=records, **options)
    try: 
        prepared_rendered = rendered.strip('\n ').encode('utf-8').decode('utf-8')
        if not options.get('full'):
            return prepared_rendered
        prettyhtml = minidom.parseString(prepared_rendered).toprettyxml(indent="  ")
        return prettyhtml
    except Exception as e:
        print(f"Error parsing template: {e}")
        return rendered
    
# End of special templates


#############################################################################################################################

class FormElement:

    options = {
        'C': 'Text',
        'N': 'Number',
        'D': 'Date',
        'L': 'Logical',
        'O': 'Options'
    }

    def __init__(self, description: str, elemtype:str="C", options:str="",
                 elemname:str="subitem", elemvalue:str="", elemid:int=0,  
                 p_class:str="renglon", label_class:str="large steelblue",
                 input_class:str="small"):
        self.description = description
        self.elemtype = "C"
        if elemtype in self.options.keys():
            self.elemtype = elemtype
        else:
            raise ValueError("Element type must be one of 'C, N, D, L, O'")
        self.options = options
        self.name = elemname
        self.id = elemid
        self.value = elemvalue
        self.p_class = p_class
        self.label_class = label_class
        self.input_class = input_class

    
    def __repr__(self):
        instance_name = f"{self.name}-{self.id}"
        if self.elemtype == 'C':
            return f'''<input value="{self.value or ''}" style="font-size: 1.3em;" type="text" id="{instance_name}" name="{instance_name}" class="{self.input_class}"/>'''
        elif self.elemtype == 'N':
            return f'''<input value="{self.value or ''}" style="font-size: 1.3em; text-align: right;" type="number" id="{instance_name}" name="{instance_name}" style="text-align: right;" />'''
        elif self.elemtype == 'D':
            return f'''<input value="{self.value or ''}" style="font-size: 1.3em;" type="date" id="{instance_name}" name="{instance_name}" />'''
        elif self.elemtype == 'L':
            return f'''
              <span style="display: inline-flex; flex-direction: row; align-items: center;">
              <strong style="font-size: 1em;"> 
                <label class="green" for="{self.description}-si">Sí</label>&nbsp;<input {'checked ' if str(self.value) == "1" else ' '}  id="{instance_name}-1" name="{instance_name}" style="transform: scale(2); margin: 5px;" type="radio" name="{self.description}" id="{self.description}-si" value="1">
                <label class="red" for="{self.description}-no">No</label>&nbsp;<input {'checked ' if str(self.value) == "0" else ' '} id="{instance_name}-0" name="{instance_name}" style="transform: scale(2); margin: 5px;" type="radio" name="{self.description}" id="{self.description}-no" value="0">
              </strong>
              </span>
            '''
        else:
            if not self.options:
                return f'<select class="{self.input_class}" style="font-size: 1.3em;" name="{self.description}">\n\t<option selected disabled value="">Debe proveer una opción.</option>\n</select>'
            else:
                options = re.split(r"\s*,\s*", self.options)
                optionstring = "\n".join([f'''\t<option {' selected ' if o == self.value else ' '} value="{o}">{o.title()}</option>''' for o in options])
                return f'''<select class="{self.input_class}"  id="{instance_name}" name="{instance_name}" style="font-size: 1.2em; padding: 5px; padding-left: 10px; padding-right: 10px;">\n{optionstring}\n</select>'''

    def __str__(self):
        response = f'''
        <p class="{self.p_class}">
          <label class="{self.label_class}">{self.description}</label>
          {repr(self)}
        </p>
        '''
        return response

#############################################################################################################################

def test1():
    import asyncio
    from server import run_dev_server
    from app import OxiApp

    head = Head("Tina Demo App")
    head.add_item('script', '<script>console.clear(); console.info("Tina Demo App loaded.");</script>')
    menubar = NavBar("TinA Demo App", [MenuItem("Home", "/"),
                       DropdownMenu("Text", [MenuItem("A", "/a"), MenuItem("B", "/b")]),
                       DropdownMenu("Numbers", [MenuItem("One", "/one"), MenuItem("Two", "/two")]),
                       DropdownMenu("Menu Colors", [MenuItem("Normal", "/colors"), MenuItem("Reversed", "/colors?reverse")]),
                       MenuItem("About", "/about")
                       ])
    # menubar.mB = '#008f00'
    menubar.mB = '#808080'
    htmlApp = HTMLApp(head=head, body=Body(menubar))

    app = OxiApp()

    @app.get('/')
    async def home():
        return render_template(str(htmlApp), contents="<p></p><p></p><h2 style='color: steelblue; shadow: 3px;'>Home</h2>")

    @app.get('/a')
    async def a():
        return render_template(str(htmlApp), contents="<p></p><p></p><h2 style='color: red;'>A</h2>")

    @app.get('/b')
    async def b():
        return render_template(str(htmlApp), contents="<p></p><p></p><h2 style='color: green;'>B</h2>")

    @app.get('/one')
    async def one():
        return render_template(str(htmlApp), contents="<p></p><p></p><h2 style='color: magenta;'>One</h2>")

    @app.get('/two')
    async def two():
        return render_template(str(htmlApp), contents="<p></p><p></p><h2 style='color: silver;'>Two</h2>")

    @app.get('/colors')
    async def colors():

        def reversestr(s):
            l = list(s)
            l.reverse()
            return ''.join(l)


        template_filters_bak = TemplateLight.template_filters.copy()

        TemplateLight.template_filters['title'] = str.title
        TemplateLight.template_filters['capitalize'] = str.capitalize
        TemplateLight.template_filters['upper'] = str.upper
        TemplateLight.template_filters['lower'] = str.lower
        TemplateLight.template_filters['reverse'] = reversestr
        TemplateLight.template_filters['toJson'] = lambda x: json.dumps(x, ensure_ascii=False)

        retval = f"<p></p><p></p><h1 style='color: {random.choice(['red', 'green', 'blue'])}; text-align: center;'>Colors</h1>"
        retval += "<p>&nbsp;</p><p>&nbsp;</p>"
        retval += '<div>'

        reversefilter = True if b'reverse' in app.scope.get('query_string') else False

        template = '''
        {% for k, v in items %}
            <div 
              style="display:flex; 
                flex-direction: row; 
                justify-content: space-around; 
                align-items: center; 
                font-size: 200%; 
                margin-bottom: 15px;"
            >
                <span style="width: 10em; max-width: 10em; min-width: 10em; font-weight: bodld;">
                {% if rf %}
                    {{ k|upper|reverse }}
                {% else %}
                    {{ k|upper }}
                {% endif %}
                </span>
                <span 
                  style="width: 1.5em, min-width: 1.5em, max-width: 1.5em; height: 1.5em; min-height: 1.5em; 
                  max-height: 1.5em; border: solid 1px; background: {{ v }};"
                >
                    &nbsp;&nbsp;&nbsp;&nbsp;&nbsp;
                </span>
            </div>
        {% endfor %}
        '''

        retval += render_template(template, items=MenuColors.items(), rf=reversefilter)            
        retval += '</div>'
        rendered = render_template(str(htmlApp), contents=retval)
        TemplateLight.template_filters = template_filters_bak.copy()
        return rendered
    
    @app.get('/about')
    async def about():
        html="<p></p><p></p><h2 style='color: steelblue;'>Every app needs an 'about' page. This is it.</h2>"
        return render_template(str(htmlApp), contents=html)


    print("\nServer HTML content at port 8410.")
    asyncio.run(run_dev_server(app, '', 8410))

def test2():
    import sys, subprocess
    tpl = """
Hi, I am {{ user }}!
"""
    # pre = preprocess_template(tpl)
    subprocess.run(("clear",))
    usr = sys.argv[1] if len(sys.argv) > 1 else "Fico"
    src = sys.argv[2] if len(sys.argv) > 2 else "/nowhere/never"
    rendered = render_template(tpl, user= usr, source=src)
    print(rendered)

if __name__ == '__main__':
    try:
        test1()
    except KeyboardInterrupt:
        print("\nServer is going down.\nBye, bye...")
    except Exception as exc:
        print(f"\nUnexpected error: {exc}\nQuitting...")



# -*- coding: utf-8 -*-

from typing import Callable

try:
    from . import  __version__ as oxi_version, __oxi_port__ as oxi_port, __oxi_host__ as oxi_host
    from .utils import is_windows, is_linux, is_mac, to_bytes
    from .config import Config
except ImportError:
    from activate_this import oxi_env
    if not oxi_env:
        raise ImportError("Oxi environment not activated. Please activate the virtual environment.")
    from oxi import  __version__ as oxi_version, __oxi_port__ as oxi_port, __oxi_host__ as oxi_host
    from oxi.utils import is_linux, is_windows, is_mac
    from oxi.config import Config

oxi_version

async def static_app(scope: dict, receive: Callable, send: Callable, absolute_path= False)-> None:

    remote_host = scope.get('remote-host')
    # remote_ip = scope.get('x-real-ip')
    remote_ip = scope.get('client', ['unknown','unknown'])[0]
    method = scope.get('method').upper()
    pth = scope.get('path')
    if pth.endswith("/") and len(pth) > 1:
        pth = pth[:-1]

    async def send_error(code, resource = None):
        statusobj = HTTPStatus(code)
        template = TinaHandler.get_error_template()
        resource = '' if resource is None else str(resource)
        body = template.format(statusobj.value, statusobj.value, resource, statusobj.phrase, tina_version)
        body_len = len(body)
        print(f"\n(PID {os.getpid()}) {method} {pth} request from {remote_ip}({remote_host}) {time.strftime('%Y-%m-%d %H:%M:%S')} - {statusobj.value}")
        
        await send({
        "type": "http.response.start",
        "status": statusobj.value,
        "headers": [
                [b"content-type", b"text/html; charset=utf-8"],
                [b"content-length", str(body_len).encode('utf-8')],
            ],
            })
        
        return await send({
            "type": "http.response.body",
            "body": to_bytes(body),
            "more_body": False
        })

    async def send_redirection(where, code = 302):
        statusobj = HTTPStatus(code)
        if type(where) == str:
            where = where.encode('utf-8')

        await send({
        "type": "http.response.start",
        "status": statusobj.value,
        "headers": [(b"location", where)]
            })
        return await send({
            "type": "http.response.body",
            "body": b'',
            "more_body": False
        })

    async def send_proxy_pass(key, where):
        l = reduce(lambda r, item: [*r, item] if (item and item != key) else r, pth.split("/"), [])
        newpth = "/".join(l)
        # print(newpth)
        fullpath = "/".join((where, newpth))
        if len(scope.get('query_string')):
            fullpath += "?" + scope.get("query_string").decode('latin-1')
        parsed = urlparse(fullpath)
        query_string = f"?{parsed.query}" if len(parsed.query) else ''
        request_path = parsed.path + query_string 
        request_line = f"{scope.get('method').upper()} {request_path} {scope.get('scheme').upper()}/{scope.get('http_version')}\r\n"
        headers = ""
        for k, v in scope.get('headers'):
            header = f"{k.decode()}: {v.decode()}\r\n"
            # print(header)
            headers += header

        request = request_line + headers + "\r\n"

        request_body = b''
        if scope.get("method").lower() not in [b'get', b'head', b'options']:
            if len(scope.get('request_body')):
                request_body  = scope.get('request_body') + b'\r\n'
            else:
                msg = await receive()
                request_body = msg.get('body') + b"\r\n"

        addrparts = parsed.netloc.split(":")
        addr, port = addrparts[0], "80" 
        if len(addrparts) > 1: 
            port = addrparts[1]
        addr = addr.strip()
        port = int(port.strip())

        pt_reader = None
        pt_writer = None
        try:
            pt_reader, pt_writer = await asyncio.open_connection(addr, port, limit = 65536 * 4)
            print(f"CONNECTION OPENED WITH {addr}:{port} AT {time.strftime(strftime_template)}")
        except:
            await send_error(404, pth)
            return 404

        try:
            pt_writer.write(to_bytes(request))
            await pt_writer.drain()
        except Exception as exc:
            print(f"Exception occurred while writing headers to proxy pass: '{exc}'")
            return await send_error(500, repr(exc))
        if len(request_body) > 2:
            try:
                pt_writer.write(request_body)
                await pt_writer.drain()
            except Exception as exc:
                print(f"Exception occurred while writing request body to proxy pass: '{exc}'")
                return await send_error(500, repr(exc))
        print(f"REQUEST SENT AT {time.strftime(strftime_template)}, AWAITING RESPONSE.")

        response_head = await pt_reader.readuntil(b'\r\n\r\n')
        print(f"GOT RESPONSE HEAD  AT {time.strftime(strftime_template)}")

        lines = response_head.split(b'\r\n')
        try: 
            # response_method_n_version, response_status, _ = lines[0].split(b" ")
            three_values_expected = lines[0].split(b" ")
            # response_status = int(response_status)
            response_status = int(three_values_expected[1])
            print(f"RESPONSE STATUS CODE: {response_status}")
        except Exception as exc:
            print(f"Error processing line: '{lines[0]}'")
            await send_error(500, repr(exc))
            return 500
        
        response_headers = []
        for line in lines[1:]:
            print(f"Processing response header line: '{line.decode()}'")
            if len(line):
                try:
                    k, v = line.split(b': ')
                    response_headers.append((k.lower(), v))
                except Exception as exc:
                    return await send_error(500, f"Error while processing line: '{line}': {exc}".encode('utf-8'))


        response_body = b''
        response_chunk = b''
        if response_status != 304:
            while True:
                response_chunk = await pt_reader.read()
                response_body += response_chunk
                if not len(response_chunk):
                    break
        print(f"GOT RESPONSE BODY AT {time.strftime(strftime_template)}")

        # print(response_headers)

        response_headers_dict = dict(response_headers)
        response_headers_dict[b'content-length'] = str(len(response_body)).encode('utf-8')

        response_content_type = response_headers_dict.get('content-type')
        if response_content_type and response_content_type.startswith(b'text/html'):
            print("It is HTML informed by response content-type header.")
            response_body = TinaHandler.rewrite_urls(key, response_body)
        elif response_content_type:
            print(f"It is {response_content_type}")
        else:
            if TinaHandler.is_html(response_body):
                print("It is HTML guessed examining contents")
                response_headers_dict[b'content-type'] = b'text/html; charset=utf-8'
                response_body = TinaHandler.rewrite_urls(key, response_body)
            else:
                guessed_type = guess_type(request_path)[0]
                if guessed_type:
                    response_headers_dict[b'content-type'] = guessed_type.encode('utf-8')
                    if guessed_type.startswith("text/html"):
                        response_headers_dict[b'content-type'] += b"; charset=utf-8"
                        print("It is HTML detected by guess_type")
                        response_body = TinaHandler.rewrite_urls(key, response_body)
                else:    
                    print("Shit! We don't know what it is, and we should, in order to provide 'rewrite_urls' in case it's HTML.")
                    response_body = TinaHandler.rewrite_urls(key, response_body)

        response_headers = response_headers_dict.items()
        print(f"RESPONSE HEADERS:\n{response_headers}")

        await send({
        "type": "http.response.start",
        "status": response_status,
        "headers": response_headers
            })
        
        await send({
            "type": "http.response.body",
            # "body": pt_response,
            "body": response_body,
            "more_body": False
        })

        return response_status

    async def send_mp4(filepath):
        print(f"\n(PID {os.getpid()}) {method} {pth} request from {remote_ip}({remote_host}) {time.strftime('%Y-%m-%d %H:%M:%S')} - 200")

        def getrange(rangebytes: bytes) -> tuple[int, int]:
            # with open(f"log/ranges.log", "ba") as fd:
            #     fd.write(to_bytes(pth) + b" (" + time.strftime("%X").encode() + b"): " + rangebytes + b'\n')
            _, points = rangebytes.split(b"=")
            if b"," in points:
                print(f"\n@Range requested@\n{points}\n@@@\n")
            begin, end = points.split(b"-")
            try:
                b = int(begin)
                e = int(end) if len(end) else 0
            except:
                b, e = 0, 0
            return b, e
        
        mp4 = Mp4(filepath)
        status_code = 200
        received, begin, end = 0, 0, mp4.filesize
        response_headers = TinaHandler.common_response_headers() + TinaHandler.resource_response_headers(filepath)
        bytez = TinaHandler.get_request_header(scope, b"range")
        if bytez:
            begin, end = getrange(bytez)
            end = (mp4.filesize - 1) if not end else end
            response_headers.append([b"content-range",  f"bytes {begin}-{end}/{mp4.filesize}".encode('utf-8')])
            for h in response_headers:
                if h[0] == 'content-length':
                    h[1] = str(end - begin + 1).encode('utf-8')
                    break
            status_code = 206

        stream = mp4.stream_range(begin, end)

        startdict = SmartDict(type='http.response.start', status=status_code, headers=response_headers)
        try:        
            await send(startdict)
            await send(SmartDict(type='http.response.body', body=b'', more_body=True))
            for chunk in stream:
                bodydict = SmartDict(type='http.response.body', body=chunk, more_body=True)
                await send(bodydict)
        except Exception as exc:
            print(f"Exception occurred while streaming mp4 content to client: {exc}")
        finally:
            try:
                bodydict = SmartDict(type='http.response.body', body=b'', more_body=False)
                await send(bodydict)
            except:
                pass
            print("Video stream closed.")
            return

    async def send_chunked_file(filepath):
        ftype = guess_type(filepath)
        ftype = ftype[0] if ftype and ftype[0] else "application/x-trash"
        if ftype.startswith("text"):
            ftype += "; charset=utf-8"

        def can_gzip()-> bool:
            retval: bool = not ftype.startswith('video')
            retval = retval and not ftype.endswith('jpg')
            retval = retval and not ftype.endswith('jpeg')
            ae = TinaHandler.get_request_header(scope, b'accept-encoding')
            ag = False if not ae else ae.find(b'gzip') != -1
            # ae and print(f"Request accepts encoding: {ae.decode('utf-8')}")
            retval = retval and ag

            return retval
            # return False

        print(f"\n(PID {os.getpid()}) {method} {pth} request from {remote_ip}({remote_host}) {time.strftime('%Y-%m-%d %H:%M:%S')} - 200")

        if ftype in ["video/mp4", "video/quicktime"]:
            return await send_mp4(filepath)

        ### Added 21/06/2024
        body = b''
        file_stat = os.stat(filepath)
        body_len = file_stat.st_size
        # body = b''

        file_desc = open(filepath, "rb")
        if can_gzip():
            body = gzip.compress(file_desc.read())    
            body_len = len(body)
            body = io.BytesIO(body)
        else:
            body = file_desc

        chunks, tail = TinaHandler.chunks(body_len)
        response_headers = [
                    [b"content-type", to_bytes(ftype)],
                    [b"content-length", str(body_len).encode('utf-8')],
                    [b'transfer-encoding', b'chunked']
                ]
        if can_gzip():
            response_headers = [*response_headers, [b"content-encoding",  b"gzip"]]
        if ftype.startswith('video') or ftype.startswith('audio'):
            # response_headers = [*response_headers, [b"accept-ranges",  b"bytes"], [b"connection", b"keep-alive"]]
            response_headers = [*response_headers, *TinaHandler.common_response_headers()]

        ranges = TinaHandler.get_request_header(scope, b"range")
        if ranges:
            stri = f"Ranges requested: {ranges.decode('utf-8')}"
            with open("log/file_log.txt", "a") as logfile:
                logfile.writelines([stri + '\n'])
            print(stri)

        try:
            await send({
            "type": "http.response.start",
            "status": 200,
            "headers": response_headers,
            })

            if len(chunks):
                chunk_size = TinaHandler.config.get("chunk-size")
                for index, chunk in enumerate(chunks):
                    # print(f"Sending chunk number {index + 1} ({chunk[0]}-{chunk[1]}) of {len(chunks)}")
                    print(".", end="")
                    await send({
                        "type": "http.response.body",
                        # "body": body[chunk[0]:chunk[1]],
                        "body": hex(chunk_size).lstrip('0x').encode('utf-8') + b"\r\n" + body.read(chunk_size) + b"\r\n",
                        # "body": body.read(chunk_size)
                        "more_body": True
                    })
            if tail:
                print(f"\nSending last chunk (tail) ({tail[0]}-{tail[1]}) for file {scope.get('path')}")
                tail_bytes = body.read()
                tail_len = len(tail_bytes)
                await send({
                    "type": "http.response.body",
                    # "body": body[tail[0]:tail[1]] if tail else b'',
                    "body": hex(tail_len).lstrip('0x').encode('utf-8') + b"\r\n" + tail_bytes + b"\r\n",
                    # "body": tail_bytes,
                    "more_body": True
                })
            else:
                print(f"Closing transmission for file {scope.get('path')} with no further data.")
            
            await send({
                "type": "http.response.body",
                # "body": b'',
                "body": b'0\r\n' + b'' + b'\r\n',
                # "body": b'0\r\n',
                "more_body": False
            })

        except BrokenPipeError as bpErr:
            print(f"Broken pipe ({errno.EPIPE}). Leaving intent of sending {scope.get('path')}.")
            return
        except Exception as exc:
            print(f"Exception ({exc}). Leaving intent of sending {scope.get('path')}.")
            return

    async def send_file(filepath):
        ftype = guess_type(filepath)
        ftype = ftype[0] if ftype and ftype[0] else "application/x-trash"
        if ftype.startswith("text"):
            ftype += "; charset=utf-8"

        print(f"\n(PID {os.getpid()}) {method} {pth} request from {remote_ip}({remote_host}) {time.strftime('%Y-%m-%d %H:%M:%S')} - 200")

        if ftype in ["video/mp4", "video/quicktime"]:
            return await send_mp4(filepath)

        body = b''

        def file_data():
            # file_desc = open(filepath, "rb")
            # file_stat = os.stat(filepath)
            file_desc = os.open(filepath, os.O_RDONLY | os.O_NONBLOCK)
            file_stat = os.fstat(file_desc)
            body_len = file_stat.st_size
            return body_len, file_desc
        body_len, file_desc = await asyncio.to_thread(file_data)

        response_headers = [
                    [b"content-type", to_bytes(ftype)],
                    [b"content-length", str(body_len).encode('utf-8')]
                ]
        if ftype.startswith('video') or ftype.startswith('audio'):
            response_headers = [*response_headers, *TinaHandler.common_response_headers()]

        try:
            await send({
            "type": "http.response.start",
            "status": 200,
            "headers": response_headers,
            })

            chunk_size = TinaHandler.config.get('chunk-size', 8192)
            # chunk_size = 8192
            remaining = body_len

            while remaining > 0:
                try:
                    part = os.read(file_desc, min(remaining, chunk_size))
                    if not part:
                        # raise RuntimeError(f"Something was wrong while reading {scope.get('path')}") 
                        break
                    remaining -= len(part)
                    if not remaining:
                        print(f"Finished reading {scope.get('path', '---')}")
                    await send({
                        "type": "http.response.body",
                        "body": part, 
                        "more_body": True
                    })
                except BlockingIOError as e:
                    if e.errno == errno.EAGAIN:
                        # No data available *right now*, try again later
                        # print("No data ready yet, try again")
                        asyncio.sleep(0)
                        continue
                    else:
                        raise                

            await send({
                "type": "http.response.body",
                "body": b'',
                "more_body": False
            })

        except BrokenPipeError as bpErr:
            print(f"Broken pipe ({errno.EPIPE}). Leaving intent of sending {scope.get('path')}.")
            return
        except Exception as exc:
            print(f"Exception ({exc}). Leaving intent of sending {scope.get('path')}.")
            return

    async def send_directory(path, dirpath):
        if not TinaHandler.config.get('dirlisting-enabled'):
            return await send_error(403)
        
        def get_file_details(entry):
            fullpath = dirpath + os.path.sep + entry
            size = os.path.getsize(fullpath)
            ret = f'<span>&nbsp;</span><span style="margin-left: 3em; margin-right: 8em; text-align: right;" class="black floatright">{size:,} bytes.</span>'
            return ret
        
        prevdir = os.path.split(path)[0]
        body = f"""
            <!DOCTYPE html>
            <html><head><title>Directory Listing</title>
            {TinaHandler.get_default_style()}
            </head><body style="margin-left: 1em; margin-right: 1em;">
            <h2>Directory listing for <span class="green">.{path}</span></h2></hr>
            <hr>
            <p style="margin-bottom: 1em; text-align: center; font-family: Times New Roman; font-size: 16px;">TinA/{tina_version}</p>
            <hr>
            <ul>
                <li><a title="Home" href="/">.</a></li>
                <li><a title="{prevdir}" href="{prevdir}">..</a></li>
        """
        entries = os.listdir(dirpath)
        entries.sort()
        entries.sort(key=lambda e: os.path.isdir(dirpath + os.path.sep + e), reverse=True)
        for entry in entries:
            body += f'''
            <li>
                <a href="{path + (os.path.sep if path != '/' else '') + entry}">{entry}&nbsp;
                {'<span class="silver">[DIR]</span>' if (await asyncio.to_thread(os.path.isdir, dirpath + os.path.sep + entry)) else get_file_details(entry)}</a>
            </li>'''
        body += "</ul></body></html>"
        body_len = len(body)
        print(f"\n(PID {os.getpid()}) {method} {pth} request from {remote_ip}({remote_host}) {time.strftime('%Y-%m-%d %H:%M:%S')} - 200")
        
        await send({
        "type": "http.response.start",
        "status": 200,
        "headers": [
                [b"content-type", b"text/html; charset=utf-8"],
                [b"content-length", str(body_len).encode('utf-8')],
            ],
            })
        
        return await send({
            "type": "http.response.body",
            "body": to_bytes(body),
            "more_body": False
        })
    
    async def send_cgi(cgipath):
        async def actually_send_cgi():
            environ = {}
            environ['GATEWAY_INTERFACE'] = f'CGI/{scope.get("http_version")}'
            environ['SERVER_SOFTWARE'] = scope.get("server_software")
            environ['REQUEST_METHOD'] = scope.get("method").upper()
            environ['CONTENT_LENGTH'] =  str(len(request_body))
            environ['SERVER_NAME'] = scope.get('server_name')
            environ['SERVER_PORT'] = scope.get('local_port')
            environ['REMOTE_HOST'] = scope.get('remote-host')
            environ['REMOTE_PORT'] = scope.get('remote-port')
            environ['QUERY_STRING'] = scope.get('query_string').decode('utf-8')
            environ['CGI_PATH'] = scope.get('path')
            environ['SCRIPT'] = os.path.split(scope.get('path'))[1]
            environ['X-REAL-IP'] = scope.get('x-real-ip')

            for [k, v] in scope.get("headers"):
                key = k.decode('utf-8').upper().replace('-', '_')
                if key == 'CONTENT_TYPE':
                    environ['CONTENT_TYPE'] = v.decode('utf-8')
                #print(f"Setting HTTP_{key} to {v.decode('utf-8')}")
                environ[f'HTTP_{key}'] = v.decode('utf-8')

            pipe = asyncio.subprocess.PIPE
            process = await asyncio.create_subprocess_exec(cgipath, stdin=pipe, stdout=pipe, stderr=pipe, env=environ)
            out, _ = await process.communicate(request_body)
            io_out = io.BytesIO(out)
            resp_line = io_out.readline()
            io_out.seek(0)
            resp_line = resp_line.strip(b'\r\n')
            code = b'404'
            try:
                _, code, phrase = resp_line.split(b' ', 3)
            except:
                pass
            print(f"\n(PID {os.getpid()}) {method} {pth} request from {remote_ip}({remote_host}) {time.strftime('%Y-%m-%d %H:%M:%S')} - {code.decode('utf-8')}")
            writer = scope.get('_writer')
            if not writer:
                raise RuntimeError("No writer socket to send response.")
            writer.write(out)
            try: 
                await writer.drain()
            except Exception as exc:
                print(f"Exception occurred while trying to send CGI data to remote host: '{exc}'")
                return await send_error(500, repr(exc))
            finally:
                try: 
                    writer.close()
                    await writer.wait_closed()
                except:
                    print(f"Exception occurred while trying to close pipe to remote host for CGI data: '{exc}'")
                    return await send_error(500, repr(exc))

# 
        if TinaHandler.config.get('cgi-enabled'):
            if await asyncio.to_thread(os.path.isfile, cgipath): 
                if await asyncio.to_thread(os.access, cgipath, os.X_OK):
                    return await actually_send_cgi()
                else:
                    print(f"\n(PID {os.getpid()}) {method} {pth} request from {remote_ip}({remote_host}) {time.strftime('%Y-%m-%d %H:%M:%S')} - 403")
                    return await send_error(403, os.path.split(cgipath)[1])
            else:
                print(f"\n(PID {os.getpid()}) {method} {pth} request from {remote_ip}({remote_host}) {time.strftime('%Y-%m-%d %H:%M:%S')} - 403")
                return await send_error(403,  os.path.split(cgipath)[1])
        else:
            print(f"\n(PID {os.getpid()}) {method} {pth} request from {remote_ip}({remote_host}) {time.strftime('%Y-%m-%d %H:%M:%S')} - 403")
            return await send_error(403,  os.path.split(cgipath)[1])

    body = ""     
    static_dir = TinaHandler.config.get('static-dir', 'static')
    if not static_dir or (not os.path.exists(static_dir)) or absolute_path:
        # static_dir = os.path.split(os.getcwd())[1]
        static_dir = "."
    # print(f"static directory is: {static_dir}")
    cgi_dir = TinaHandler.config.get('cgi-dir')
    # print(f"CGI directory is: {cgi_dir}")
    pth = (scope and scope.get('path')) or ''
    # print(f"Requested path is: {pth}")
    fullpath = ""
    is_cgi = False 



    if TinaHandler.config.get('redirects') and pth[1:] in TinaHandler.config.get('redirects').keys():
        d = TinaHandler.config.get('redirects')
        where = d.get(pth[1:])
        print(f"(PID {os.getpid()}) \nRedirect {method} {pth} request from {remote_ip}({remote_host}) {time.strftime('%Y-%m-%d %H:%M:%S')} - 302")
        return await send_redirection(where)

#    if pth[1:] in TinaHandler.config.get('proxy_pass').keys():
    for key in TinaHandler.config.get('proxy_pass').keys():
        if pth[1:].startswith(key):
            d = TinaHandler.config.get('proxy_pass')
            # where = d.get(pth[1:])
            where = d.get(key)
            retcode = await send_proxy_pass(key, where)
            print(f"\n(PID {os.getpid()}) Proxy pass {method} {pth} request from {remote_ip}({remote_host}) {time.strftime('%Y-%m-%d %H:%M:%S')} - {retcode}")
            return 
    
    if pth in easter_eggs.keys() and scope.get('method').lower().strip() == 'get':
        body = TinaHandler.get_default_style() + '<div>'
        body += easter_eggs[pth].replace("\n", "<br/>") + "</div><p>&nbsp;</p>\r\n"
        body_len = len(body)

        await send({
        "type": "http.response.start",
        "status": 200,
        "headers": [
                [b"content-type", b"text/html; charset=utf-8"],
                [b"content-length", str(body_len).encode('utf-8')],
            ],
            })
        
        return await send({
            "type": "http.response.body",
            "body": to_bytes(body),
            "more_body": False
        })

    if pth == '/demo':
        try:
            return await demo_app(scope, receive, send)
        except Exception as exc:
            _, _, trace_back = os.sys.exc_info()
            lineno = trace_back.tb_lineno
            print(f"Exception occurred while trying to load demo app at line {lineno}: '{exc}'\nTraceback: {repr(trace_back)}")

    if pth == '/uploader':
        try:
            from formhandler import formhandler
            return await formhandler(scope, receive, send)
        except Exception as exc:
            _, _, trace_back = os.sys.exc_info()
            lineno = trace_back.tb_lineno
            print(f"Exception occurred while trying to load formhandler at line {lineno}: '{exc}'\nTraceback: {repr(trace_back)}")

    m = re.match(f".*{cgi_dir}.*", pth)
    # print(f"CGI string in position in path is: {m}")
    if m:
        is_cgi = True
        fullpath = f"{os.path.abspath('.')}{pth}"
    else:
        fullpath = f"{os.path.abspath(static_dir)}{pth}"
    # print(f"Full path is: '{fullpath}' and it is {'' if is_cgi else 'not '}CGI content.")


    if pth == '/':
        candidates = ['index.html', 'index.htm']
        for candidate in candidates:
            if await asyncio.to_thread(os.path.exists, f"{fullpath}{candidate}"):
                fullpath += candidate
                # print(f"Real fullpath: {fullpath}")
                break

    requires_body: bool = scope and scope.get('method').lower() not in ['get', 'head', 'options']
    # requires_body and print("\nThis request needs to receive body.")
    request_body = b''
    if requires_body:
        msg = await receive()
        request_body = msg.get('body')

    # Keep this just in case...    
    # form = TinaHandler.form(scope or {}, request_body)

    # len(request_body) and print(f"Request body: '{request_body.decode('latin-1')}'")

    fexists: bool = await asyncio.to_thread(os.path.exists, unquote(fullpath))
    if fexists:
        if not is_cgi:
            if await asyncio.to_thread(os.path.isdir, unquote(fullpath)):
                # print(f"Serving directory: {pth}")
                return await send_directory(pth, unquote(fullpath))
            elif await asyncio.to_thread(os.path.isfile, unquote(fullpath)):
                # print(f"Serving file: {pth}")
                return await send_file(unquote(fullpath))
        else:
            # print(f"Serving CGI: {pth}")
            return await send_cgi(unquote(fullpath))
    else:
        # print(f"{fullpath} doesn't exist.")
        return await send_error(404, pth)


class OxiApp: ...







# -- coding: utf-8 -*-

from pathlib import Path
import os
import time
import hashlib

def fs_monitor():

    print("Starting file system monitor...")
    parentdir = Path(__file__).parent

    pyfiles = [open(str(entry), 'rb') for entry in parentdir.glob("*.py")]
    hashes = [hashlib.sha1(fd.read()).hexdigest() for fd in pyfiles]


    # print(f"Watching {len(pyfiles)} files for changes.")

    while True:
        time.sleep(1)
        newpyfiles = [open(str(entry), 'rb') for entry in parentdir.glob("*.py")]
        if len(newpyfiles) != len(pyfiles):
            print(f"File count changed. Signaling server restart")
            time.sleep(0.5)  # Optional debounce
            os.execv(os.sys.executable, [os.sys.executable] + os.sys.argv)            # os.kill(os.getpid(), signal.SIGUSR1)
            # os.kill(os.getpid(), signal.SIGUSR1)
            # return
        newhashes = [hashlib.sha1(fd.read()).hexdigest() for fd in newpyfiles]
        if newhashes != hashes:
            print(f"File hashes changed. Signaling server restart")
            time.sleep(0.5)  # Optional debounce
            os.execv(os.sys.executable, [os.sys.executable] + os.sys.argv)            # os.kill(os.getpid(), signal.SIGUSR1)
            # os.kill(os.getpid(), signal.SIGUSR1)
            # return
        else:
            for fd in newpyfiles:
                fd.close()

#

if __name__ == "__main__":
    fs_monitor()


# -*- coding: utf-8 -*-

import asyncio, os, sys, random, subprocess, re, mimetypes, errno, signal, hashlib, time
from threading import Thread
from typing import Callable
from pprint import pprint
from pathlib import Path
from time import time as timestamp, strftime as tstrftime, strptime as tstrptime
from datetime import datetime as dt
from urllib.parse import urlparse, parse_qsl, unquote_plus
from functools import wraps

try:
    from . import  __version__ as oxi_version, __oxi_port__ as oxi_port, __oxi_host__ as oxi_host
    from .config import Config
    from .utils import (is_windows, is_linux, is_mac, 
                        http_status_dict as status_dict, to_bytes, no_ctrlc_echo)
    from .mp4parser import Mp4
except ImportError:
    from activate_this import oxi_env
    if not oxi_env:
        raise ImportError("Oxi environment not activated. Please activate the virtual environment.")
    from oxi import  __version__ as oxi_version, __oxi_port__ as oxi_port, __oxi_host__ as oxi_host
    from oxi.config import Config
    from oxi.utils import (is_linux, is_windows, is_mac, 
                           http_status_dict as status_dict, to_bytes, no_ctrlc_echo)
    from oxi.mp4parser import Mp4

server_software =f"Oxi/{oxi_version}"

async def is_static(resource:str) -> bool:
    """
    Check if the resource exists in the file system.
    """
    return await asyncio.to_thread(os.path.exists, resource)

async def is_file(resource:str) -> bool:
    """
    Check if the resource is a regular file.
    """
    exists = await asyncio.to_thread(os.path.exists, resource)
    return exists and await asyncio.to_thread(os.path.isfile, resource)

async def is_dir(resource:str) -> bool:
    """
    Check if the resource is a directory.
    """
    exists = await asyncio.to_thread(os.path.exists, resource)
    return exists and await asyncio.to_thread(os.path.isdir, resource)

async def is_exe(resource:str) -> bool:
    """
    Check if the resource is an executable file.
    """
    exists = await asyncio.to_thread(os.path.exists, resource)
    return exists and (await asyncio.to_thread(os.access, resource, os.X_OK))

async def is_cgi_exe(resource:str, cgi_dir:str="cig-bin") -> bool:
    """
    Check if the resource is a CGI executable.
    """
    exists = await asyncio.to_thread(os.path.exists, resource)
    if not exists:
        return False
    is_exec = await is_exe(resource)
    if not is_exec:
        return False
    if not cgi_dir in resource:
        return False
    return True

async def finalize_writer(writer):
    try:
        # writer.write(b"\r\n\r\n")
        await writer.drain()
    except Exception as e:
        print(f"Writer drain error: {e}")
    finally:
        try:
            writer.close()
            await writer.wait_closed()
        except Exception as e:
            print(f"Writer close error: {e}")

enctypes: tuple = (b"application/x-www-form-urlencoded", 
                   b"multipart/form-data", 
                   b"application/json",
                   b"text/plain")

strftime_template = "%a, %d  %b %Y %X %Z"

original_zen = """Gur Mra bs Clguba, ol Gvz Crgref

Ornhgvshy vf orggre guna htyl.
Rkcyvpvg vf orggre guna vzcyvpvg.
Fvzcyr vf orggre guna pbzcyrk.
Pbzcyrk vf orggre guna pbzcyvpngrq.
Syng vf orggre guna arfgrq.
Fcnefr vf orggre guna qrafr.
Ernqnovyvgl pbhagf.
Fcrpvny pnfrf nera'g fcrpvny rabhtu gb oernx gur ehyrf.
Nygubhtu cenpgvpnyvgl orngf chevgl.
Reebef fubhyq arire cnff fvyragyl.
Hayrff rkcyvpvgyl fvyraprq.
Va gur snpr bs nzovthvgl, ershfr gur grzcgngvba gb thrff.
Gurer fubhyq or bar-- naq cersrenoyl bayl bar --boivbhf jnl gb qb vg.
Nygubhtu gung jnl znl abg or boivbhf ng svefg hayrff lbh'er Qhgpu.
Abj vf orggre guna arire.
Nygubhtu arire vf bsgra orggre guna *evtug* abj.
Vs gur vzcyrzragngvba vf uneq gb rkcynva, vg'f n onq vqrn.
Vs gur vzcyrzragngvba vf rnfl gb rkcynva, vg znl or n tbbq vqrn.
Anzrfcnprf ner bar ubaxvat terng vqrn -- yrg'f qb zber bs gubfr!"""

zen = """
The Zen of Python, by Tim Peters

Beautiful is better than ugly.
Explicit is better than implicit.
Simple is better than complex.
Complex is better than complicated.
Flat is better than nested.
Sparse is better than dense.
Readability counts.
Special cases aren't special enough to break the rules.
Although practicality beats purity.
Errors should never pass silently.
Unless explicitly silenced.
In the face of ambiguity, refuse the temptation to guess.
There should be one-- and preferably only one --obvious way to do it.
Although that way may not be obvious at first unless you're Dutch.
Now is better than never.
Although never is often better than *right* now.
If the implementation is hard to explain, it's a bad idea.
If the implementation is easy to explain, it may be a good idea.
Namespaces are one honking great idea -- let's do more of those!
"""


class ProtocolFactory: 
    """
    ProtocolFactory class to handle the server protocol.
    """
    
    success_line = "HTTP/1.1 200 OK\r\n"
    partial_line = "HTTP/1.1 206 Partial Content\r\n"
    bad_request_line = "HTTP/1.1 400 Bad Request\r\n"
    not_implemented_line = "HTTP/1.1 501 Not Implemented\r\n"
    forbidden_line = "HTTP/1.1 403 Forbidden\r\n"
    moved_line = "HTTP/1.1 301 Moved Permanently\r\n"
    moved_temporary_line = "HTTP/1.1 302 Moved Temporarily\r\n"
    error_line = "HTTP/1.1 500 Internal Server Error\r\n"
    not_found_line = "HTTP/1.1 404 Not Found\r\n"

    @property
    def base_dir(self) -> str:
        if not os.path.exists(self._base_dir):
            self._base_dir = ''
        return f"{os.path.sep}{self._base_dir}"

    @base_dir.setter
    def base_dir(self, value: str) -> None:
        if os.path.exists(value):
            self._base_dir = value
        else:
            raise ValueError(f"Base directory '{value}' does not exist.")

    @property
    def full_base_dir(self) -> str:
        return f"{self.cwd}{self.base_dir}"

    async def has_index(self):
        candidates = self.index_files if hasattr(self, 'index_files') else ['index.html', 'index.htm']
        full_candidates = [os.path.join(self.full_base_dir, f) for f in candidates]
        for candidate in full_candidates:
            if await is_file(candidate):
                return True, candidate
        return False, None
    
    def __init__(self, app: Callable = None, *, base_dir: str = None):
        self.app = app
        if self.app:
            self.app._protocol = self
        self.cwd = os.getcwd()

        for k in Config.keys():
            setattr(self, k, Config[k])
        
        self._base_dir = base_dir or (self.static_dir if hasattr(self, 'static_dir') else 'static')
        self.version = oxi_version
        self.zen = zen
        self.original_zen = original_zen
        self.strftemplate = strftime_template
        self.enctypes = enctypes

    async def __call__(self, reader: asyncio.StreamReader, writer: asyncio.StreamWriter):
        """
        Handle the incoming request and send a response.
        """
        try:
            method, path, protocol = await self.get_request_line(reader)
            parsed = urlparse(path)
            path = parsed.path
            query = parsed.query
            if query:
                query = dict(parse_qsl(query))
            else:
                query = {}
            netloc = parsed.netloc

        except ValueError as e:
            print(f"Error parsing request line: {e}")
            return await self.send_status_response(writer, status_code=400, msg=str(e))
        print(f"Received request: {method} {path} {protocol}")
        if method not in ["HEAD", "GET", "POST", "PUT", "PATCH", "DELETE"]:
            msg = f"Method {method} not allowed."
            print(msg)
            return await self.send_status_response(writer, status_code=405, msg=msg)

        try:
            request_headers_block= await self.get_headers_block(reader)
            # print(f"\nHeaders block:\n{request_headers_block.decode('utf-8')}\n\n")
        except ValueError as e:
            print(f"Error parsing  request headers: {e}")
            return await self.send_status_response(writer, status_code=400, msg=str(e))
        request_headers = await self.parse_headers(request_headers_block)
        if path == "/oxiserver_demo" and method == "GET":
            print(f"Serving Oxi Server demo page.")
            return await self.oxiserver_demo(writer=writer)
        else:
            fullpath = os.path.join(self.full_base_dir, path.lstrip("/").replace("/", os.path.sep)) 
            if await is_file(fullpath):
                if not await is_cgi_exe(fullpath, self.cgi_dir):
                    if method == "GET":
                        print(f"Serving file: {fullpath}")
                        return await self.send_file(writer=writer, fullpath=fullpath, headers=request_headers)
                    else:
                        print(f"Method {method} not allowed for file: {fullpath}")
                        return await self.send_status_response(writer, status_code=405, msg=f"Method {method} not allowed for file.")  
                else:
                    print(f"Serving CGI executable: {fullpath}")
                    return await self.send_file(writer=writer, fullpath=fullpath, headers=request_headers)
            elif path == "/":
                exists, index_file = await self.has_index()
                if exists:
                    print(f"Serving index file: {index_file}")
                    return await self.send_file(writer=writer, fullpath=index_file, headers=request_headers)
                else:
                    if not self.allow_dirlisting:
                        print(f"Directory listing not allowed: {fullpath}")
                        return await self.send_status_response(writer, status_code=403, msg="Directory listing not allowed.")
                    print(f"Serving static directory: {self.static_dir}")
                    return await self.send_directory(writer, path=path, dirpath=self.full_base_dir)
            elif await is_dir(fullpath):
                if not self.allow_dirlisting:
                    print(f"Directory listing not allowed: {fullpath}")
                    return await self.send_status_response(writer, status_code=403, msg="Directory listing not allowed.")
                print(f"Serving directory: {fullpath}")
                return await self.send_directory(writer, path=path, dirpath=fullpath)
            else:
                if self.app is not None:
                    print(f"Serving app: {self.app.name}")
                    return await self.app(reader=reader, writer=writer)
                else:
                    print(f"File {path} not found.")
                    return await self.send_status_response(writer, status_code=404, msg=f"Resource '{path}' not found.")  

    @classmethod
    async def send_directory(cls, writer: asyncio.StreamWriter, path: str, dirpath: str):
        
        async def get_file_details(entry):
            fullpath = dirpath + os.path.sep + entry
            size = await asyncio.to_thread(os.path.getsize,fullpath)
            ret = f'<span style="text-align: right;" class="black">{size:,} bytes</span>'
            return ret
        
        pth = Path(path.strip('/'))
        prevdir = ("/" + str(pth.parent)) if path != '/' else '/'
        body = f"""
            <!DOCTYPE html>
            <html><head><title>Directory Listing</title>
            <style>
                body {{
                    background-color: #f0f0f0;
                    font-family: Helvetica, Arial, sans-serif;
                    font-size: 16px;
                }}
                .green {{
                    color: green;
                }}
                .silver {{
                    color: silver;
                }}
                .black {{
                    color: black;
                }}
                .renglon_dirlist {{
                    margin-left: 1em;
                    width: 50%;
                    max-width: 50%;
                    display: flex;
                    flex-direction: row;
                    justify-content: space-between;
                    align-items: center;
                    list-style-type: none;
                    font-size: 150%;
                    margin-bottom: 0.5em;
                }}
                a.link {{
                    text-decoration: none;
                    color: steelblue;
                    font-weight: bold;
                }}
                a.link:hover {{
                    color: black;
                }}
                @media (max-width: 798px) {{
                    .renglon_dirlist {{
                        width: 90%;    
                        max-width: 90%;
                    }}
                }}
            </style>
            </head><body style="margin-left: 1em; margin-right: 1em;">
            <h2>Directory listing for <span class="green">.{str(path)}</span></h2></hr>
            <hr>
            <p style="margin-bottom: 1em; text-align: center; font-family: Times New Roman; font-size: 16px;">{server_software}</p>
            <hr>
            <ul>
                <li class="renglon_dirlist"><a class="link" title="Home" href="/">.</a></li>
                <li class="renglon_dirlist"><a class="link" title="{str(prevdir)}" href="{str(prevdir)}">..</a></li>
        """
        entries = await asyncio.to_thread(os.listdir, dirpath)
        entries.sort()
        entries.sort(key=lambda e: os.path.isdir(dirpath + os.path.sep + e), reverse=True)
        for entry in entries:
            body += f'''
            <li class="renglon_dirlist">
                <a class="link" href="{path + (os.path.sep if path != '/' else '') + entry}">{entry}</a>
                {'<span class="silver">[DIR]</span>' if (await asyncio.to_thread(os.path.isdir, dirpath + os.path.sep + entry)) else (await get_file_details(entry))}
            </li>'''
        body += "</ul></body></html>"
        body_len = len(body)
        
        # print(f"\n(PID {os.getpid()}) {method} {pth} request from {remote_ip}({remote_host}) {time.strftime('%Y-%m-%d %H:%M:%S')} - 200")
        
        try:
            writer.write(cls.success_line.encode("utf-8"))
            writer.write(f"Server: {server_software}\r\n".encode("utf-8"))
            writer.write(b"content-type: text/html; charset=utf-8\r\n"),
            writer.write(f"content-length: {str(body_len)}\r\n".encode('utf-8'))
            await writer.drain()
        except Exception as e:
            print(f"Error writing directory listing headers: {e}")
            # return await self.send_status_response(writer, status_code=500, reason="Internal Server Error", msg=str(e))
            return
        try:
            writer.write(b"\r\n")
            await writer.drain()
            writer.write(to_bytes(body))
            await writer.drain()
        except Exception as e:
            print(f"Error directory listing writing body: {e}")
            # return await self.send_status_response(writer, status_code=500, reason="Internal Server Error", msg=str(e))

    @classmethod
    async def send_file(cls, writer: asyncio.StreamWriter, fullpath: str, 
                        headers: dict = None, forced: bool = False) -> None:

        content_type = mimetypes.guess_type(fullpath)[0] or "application/octet-stream"
        if content_type == 'video/mp4' and not forced:
            return await cls.send_mp4(writer=writer, fullpath=fullpath, headers=headers)
        file_desc = await asyncio.to_thread(os.open, fullpath, os.O_RDONLY | os.O_NONBLOCK)
        file_stat = await asyncio.to_thread(os.fstat, file_desc)
        body_len = file_stat.st_size
        try:
            writer.write(cls.success_line.encode("utf-8"))
            writer.write(f"Server: {server_software}\r\n".encode("utf-8"))
            writer.write(f"Content-Type: {content_type}\r\n".encode("utf-8"))
            writer.write(f"Content-Length: {body_len}\r\n".encode("utf-8"))
            writer.write("Access-Control-Allow-Origin: *\r\n".encode("utf-8"))
            writer.write(b"\r\n")
            await writer.drain()
        except Exception as e:
            print(f"Error writing headers: {e}")
            await asyncio.to_thread(os.close, file_desc)
            # return await self.send_status_response(writer, status_code=500, reason="Internal Server Error", msg=str(e))
            return
        
        async def send_windows():
            remaining = body_len
            while remaining > 0:
                chunk = min(cls.chunk_size, remaining)
                try:
                    data = os.read(file_desc, chunk)
                    if not data:
                        break
                    writer.write(data)
                    await writer.drain()
                    remaining -= len(data)
                except BlockingIOError as e:
                    if e.errno == errno.EAGAIN:
                        await asyncio.sleep(0)
                        continue
            await asyncio.to_thread(os.close, file_desc)
            await finalize_writer(writer)

        async def send_linux():
            loop = asyncio.get_running_loop()
            try:
                file_obj = open(file_desc, "rb", closefd=False)
                await loop.sendfile(writer.transport, file_obj)
            except (AttributeError, NotImplementedError, RuntimeError) as e:
                print(f"loop.sendfile() not available or failed ({e}), falling back to read/write")
                return await send_windows()
            await asyncio.to_thread(os.close, file_desc)
            await finalize_writer(writer)

        async def send_mac():
            sock = writer.get_extra_info("socket")
            sock_fd = sock.fileno() 
            offset = 0
            while offset < body_len:
                sent = os.sendfile(sock_fd, file_desc, offset, body_len - offset)
                if sent == 0:
                    break
                offset += sent
            await asyncio.to_thread(os.close, file_desc)
            await finalize_writer(writer)

        if is_windows():
            return await send_windows()
        
        if is_linux():
            # return await send_mac()
            return await send_linux()
        
        if is_mac():
            return await send_mac()

    @classmethod
    async def send_mp4(cls, writer: asyncio.StreamWriter, fullpath: str, headers: dict = None) -> None:
        mp4 = Mp4(fullpath)
        boundaries = await mp4.faststart_boundaries
        response_line = cls.partial_line
        start, end = 0, mp4.filesize - 1

        sec_fetch_dest = headers.get("sec-fetch-dest", None)
        if sec_fetch_dest:
            if sec_fetch_dest != "video":
                response_line = cls.success_line
        bytesrange = headers.get("range")
        if not bytesrange:
            print(f"Range header not found. Sending full file.")
            start = 0
            end = 0
            # boundaries = await mp4.faststart_boundaries
            for entry in boundaries:
                moov_bounds = entry.get('moov')
                if moov_bounds:
                    end = moov_bounds[1] - 1
                    break
            if end == 0:
                end = mp4.filesize - 1
        else:
            bytesrange = bytesrange.lstrip("bytes=")
            start, end = bytesrange.split("-")
            start = int(start)
            end = int(end) if end else mp4.filesize - 1
            if start < 0 or end >= mp4.filesize or start > end:
                print(f"Invalid range: {bytesrange}. Sending full file.")
                start, end = 0, mp4.filesize - 1
        length = end - start + 1
        try:
            writer.write(response_line.encode("utf-8"))
            writer.write(f"Server: {server_software}\r\n".encode("utf-8"))
            writer.write(b"Content-Type: video/mp4\r\n")
            # writer.write(b"Access-Control-Allow-Origin: *\r\n")
            writer.write(b"Accept-Ranges: bytes\r\n")
            print(f'\nSENDING video content:\tContent-Range: bytes {start}-{end}/{mp4.filesize}\n')
            writer.write(f'Content-Range: bytes {start}-{end}/{mp4.filesize}\r\n'.encode('utf-8'))                        
            writer.write(f"Content-Length: {length}\r\n\r\n".encode("utf-8"))
            await writer.drain()
        except Exception as e:
            print(f"Error writing mp4 headers: {e}")
            # return await cls.send_status_response(writer, status_code=500, reason="Internal Server Error", msg=str(e))
            return

        stream = mp4.async_stream_range(start, end)
        written = 0
        async for chunk in stream:
            try:
                writer.write(chunk)
                await writer.drain()
                written += len(chunk)
                # print(f"Written {len(chunk)} bytes for a total of {written} out of {mp4.filesize}.")
            except Exception as e:
                print(f"Error writing mp4 chunk: {e}")
                # return await cls.send_status_response(writer, status_code=500, reason="Internal Server Error", msg=str(e))
                break
        try:    
            writer.close()
            await writer.wait_closed()
        except Exception as e:
            print(f"Error closing writer: {e}")
            # return await cls.send_status_response(writer, status_code=500, reason="Internal Server Error", msg=str(e))
        finally:
            print(f"{written} bytes written out of {mp4.filesize}.")
            return

    @classmethod    
    async def get_request_line(cls, reader: asyncio.StreamReader) -> tuple[str]:
        """
        Read the request line from the client.
        """
        retries = 0
        max_retries = 3
        while True:
            try:
                request_line = await reader.readuntil(b"\r\n")
                break
            except asyncio.IncompleteReadError as e:
                retries += 1
                print(f"Retrying to read request line: {e}\nAttempt {retries}/{max_retries}")
                if retries >= max_retries:
                    request_line = ''
                    break
                else:
                    await asyncio.sleep(0)
                    continue
        if not request_line:
            raise ValueError("Request line is incomplete after multiple attempts.")
        request_line = request_line.rstrip(b"\r\n")
        if not request_line:
            raise ValueError("Request line is empty after stripping.")
        request_line = request_line.decode("utf-8").strip()
        if not request_line:
            raise ValueError("Request line is empty after decoding.")
        try:
            # self.request_line = request_line
            method, path, protocol = re.split(r"\s+", request_line)
        except ValueError:
            raise ValueError("Request line is malformed. Unable to split into method, path, and protocol.")
        if not method or not path or not protocol:
            raise ValueError("Request line is malformed. Missing method, path, or protocol.")
        return method.upper(), unquote_plus(path), protocol

    @classmethod
    async def get_headers_block(cls, reader: asyncio.StreamReader) -> bytes:
        """
        Read header lines from the client.
        """
        retries = 0
        max_retries = 3
        while True:
            try:
                headers_block = await reader.readuntil(b"\r\n\r\n")
                break
            except asyncio.IncompleteReadError as e:
                retries += 1
                print(f"Retrying to read header blocks: {e}\nAttempt {retries}/{max_retries}")
                if retries >= max_retries:
                    headers_block = b''
                    break
                else:
                    await asyncio.sleep(0)
                    continue
        if not headers_block:
            raise ValueError("Headers block is incomplete after multiple attempts.")
        headers_block = headers_block.rstrip(b"\r\n\r\n")
        if not headers_block:
            raise ValueError("Headers block is empty after stripping.")
        # self.headers_block = headers_block
        return headers_block

    @classmethod
    async def parse_headers(cls, headers_block: bytes) -> dict:
        """
        Parse the headers block into a dictionary.
        """
        headers = {}
        lines = headers_block.decode("utf-8").split("\r\n")
        for line in lines:
            if ": " in line:
                key, value = line.split(": ", 1)
                headers[key.strip().lower()] = value.strip()
        return headers
    
    @classmethod
    async def send_status_response(cls, writer: asyncio.StreamWriter, status_code: int=404, reason: str = None, msg:str="") -> None:
        reason = reason or status_dict.get(status_code, "Unknown Status")
        status_line = f"HTTP/1.1 {status_code} {reason}\r\n"
        realmsg = msg + '\r\n' if len(msg) else ''
        response = f"""
<html>
<head><title>{status_code} {reason}</title></head>
<body style="background-color: #f0f0f0; font-family: Helvetica, Arial, sans-serif;">
<center><h1>{status_code} {reason}</h1></center>
<hr><center>{server_software}</center>
<p>&nbsp;</p>
<p>{realmsg}</p>
</body>
</html>"""
        if writer.is_closing():
            print(f"Writer is closing. Cannot send status response.")
            return
        
        try:
            writer.write(status_line.encode("utf-8"))
            writer.write(f"Server: {server_software}\r\n".encode("utf-8"))
            writer.write(b"Content-Type: text/html; charset=utf-8\r\n")
            writer.write(f"Content-Length: {len(response)}\r\n\r\n".encode("utf-8"))
            writer.write(response.encode("utf-8"))
            await writer.drain()
        except Exception as e:
            print(f"Error writing status response: {e}")
            return
        try:
            writer.close()
            await writer.wait_closed()
        except Exception as e:
            print(f"Error closing writer: {e}")
            return

    async def oxiserver_demo(self, writer: asyncio.StreamWriter = None) -> None:
        """
        Send a demo HTML page to the client.
        """
        if writer is None:
            raise ValueError("writer is None. Please provide a valid writer.")
        # Let's begin slowly
        exists = await is_static("./static/img")
        if not exists:
            msg = "Static directory not found. Please create a static directory with images."
            return await self.send_status_response(writer, status_code=404, msg=msg)
        listdir = await asyncio.to_thread(os.listdir, "./static/img")
        listdir = [entry for entry in listdir if mimetypes.guess_type(entry)[0] and mimetypes.guess_type(entry)[0].startswith("image")]
        img_src = random.choice(listdir)
        newzen = random.choice([self.zen, self.zen, self.zen, self.original_zen]).replace("\n", "<br>")
        html = f"""
        <html>
            <head>
                <title>Oxi Server</title>
                <!--<link rel="stylesheet" href="static/css/style.css">-->
            </head>
            <body style="background-color: #f0f0f0; font-size: 120%; font-family: Helvetica, Arial, sans-serif;">
                <div style="text-align: center;">
                    <h1 style="color: steelblue;">Welcome to <strong style="color: red;">Oxi Server</strong>!</h1>
                    <img src="/img/{img_src}" alt="Random Image" width="20%" style="border-radius: 20px;">
                </div>
                <h2>Version: <strong>{self.version}</strong></h2>
                <hr/>
                <p>{newzen}</p>
                <hr>
                <p style="text-align: center;">
                    <a href="/">Home</a>
                </p>
                <hr>
                <p style="text-align: right;ont-size: 16px; margin-right 1em;">{server_software}</p>
                <script>
                    setInterval(() => location.href = location.href, 10000);
                </script>
            </body>
        </html>
"""
        content_length = len(html)
        print(f"Writing to socket {content_length} bytes.\n")
        writer.write(f"Server: {server_software}\r\n".encode("utf-8"))
        content_type = "text/html; charset=utf-8"
        writer.write(self.success_line.encode("utf-8"))
        writer.write(f"Content-Type: {content_type}\r\n".encode("utf-8"))
        writer.write(f"Content-Length: {content_length}\r\n".encode("utf-8"))
        writer.write(b"\r\n")
        writer.write(html.encode("utf-8"))
        writer.write(b"\r\n")
        await writer.drain()
        writer.close()
        await writer.wait_closed()

######################################################################################

def fs_monitor():

    parentdir = Path(__file__).parent

    pyfiles = [open(str(entry), 'rb') for entry in parentdir.glob("*.py")]
    hashes = [hashlib.sha1(fd.read()).hexdigest() for fd in pyfiles]
    for fd in pyfiles:
        fd.close()

    while True:
        time.sleep(1)
        newpyfiles = [open(str(entry), 'rb') for entry in parentdir.glob("*.py")]
        if len(newpyfiles) != len(pyfiles):
            for fd in newpyfiles:
                fd.close()
            print(f"File count changed. Signaling server restart")
            time.sleep(0.5)  # Optional debounce
            os.execv(os.sys.executable, [os.sys.executable] + os.sys.argv)            # os.kill(os.getpid(), signal.SIGUSR1)
            # os.kill(os.getpid(), signal.SIGUSR1)
            # return
        newhashes = [hashlib.sha1(fd.read()).hexdigest() for fd in newpyfiles]
        for fd in newpyfiles:
            fd.close()
        if newhashes != hashes:
            print(f"File hashes changed. Signaling server restart")
            time.sleep(0.5)  # Optional debounce
            os.execv(os.sys.executable, [os.sys.executable] + os.sys.argv)            # os.kill(os.getpid(), signal.SIGUSR1)
            # os.kill(os.getpid(), signal.SIGUSR1)
            # return

######################################################################################

async def run_dev_server(protocol: Callable = ProtocolFactory(), 
                         host:str=oxi_host, port:int=oxi_port, 
                         unix_socket:str=None) -> None:
    
    subprocess.run("clear")

    with no_ctrlc_echo():
        stop_event = asyncio.Event()
        loop = asyncio.get_running_loop()

        def _signal_handler():
            print("\nSignal received. Shutting down...")
            stop_event.set()

    
        loop.add_signal_handler(signal.SIGINT, _signal_handler)
        loop.add_signal_handler(signal.SIGTERM, _signal_handler)
        
        server = None
        if unix_socket:
            if os.path.exists(unix_socket):
                os.remove(unix_socket)
            server = await asyncio.start_unix_server(protocol, path=unix_socket, loop=loop)
        else:
            server = await asyncio.start_server(protocol, host, port, reuse_address=True, reuse_port=True)

        server_task = asyncio.create_task(server.serve_forever())

        if unix_socket:
            print(f"\n Oxi Server running at {unix_socket}\n")
        else:
            print(f"\n Oxi Server running at {host}:{port}\n")

        # Wait for shutdown signal
        await stop_event.wait()
        print("Stopping Oxi server...")

        # Cleanup
        server.close()
        await server.wait_closed()
        server_task.cancel()

        try:
            await server_task
        except asyncio.CancelledError:
            pass

        print("Oxi Server shut down cleanly.")

async def runner(port: int = oxi_port, host: str = oxi_host) -> None:
    # protocol = ProtocolFactory()
    # protocol.allow_dirlisting = False
    # await run_dev_server(protocol, host=oxi_host, port=oxi_port)
    await run_dev_server(host=host, port=port) # Uses ProtocolFactory() by default

def main():
    # Start the file system monitor in a separate thread
    fs_monitor_thread = Thread(target=fs_monitor, daemon=True)
    fs_monitor_thread.start()
    # Run the server
    port = oxi_port
    if len(sys.argv) > 1:
        try:
            port = int(sys.argv[1])
        except ValueError:
            print(f"Invalid port number '{sys.argv[1]}'. Using default port {oxi_port}.")
    asyncio.run(runner(port=port, host=oxi_host ))

if __name__ == "__main__":
    main()


import time
from contextlib import contextmanager
import sys, termios

@contextmanager
def no_ctrlc_echo():
    if not sys.stdin.isatty():
        yield
        return
    fd = sys.stdin.fileno()
    original = termios.tcgetattr(fd)
    try:
        modified = original[:]
        modified[3] &= ~termios.ECHOCTL
        termios.tcsetattr(fd, termios.TCSANOW, modified)
        yield
    finally:
        termios.tcsetattr(fd, termios.TCSANOW, original)

if __name__ == "__main__":
    try:
        with no_ctrlc_echo():
            print("Running. Press Ctrl+C to test.")
            time.sleep(30)
    except KeyboardInterrupt:
        print("Got KeyboardInterrupt.")





# -*- coding: utf-8 -*-
__version__ = "0.2.33"
__author__ = "Domingo E. Savoretti"
__email__ = "esavoretti@gmail.com"
__license__ = "MIT"
__url__ = "https://github.com/sandy98/oxi"
__description__ = "ASGI full stack: Web Server, Web App and a ton of goodies."
__copyright__ = "Copyright (c) 2025 Domingo E. Savoretti"
__oxi_port__ = 8086
__oxi_host__ = "0.0.0.0"

__all__ = [
    "utils",
    "mp4parser",
    "template",
    "websocket",
    "app",
    "server",
    "config",
]


# -*- coding: utf-8 -*-

import os, json

Config = {
    'static_dir': 'static',
    'template_dir': 'templates',
    'upload_dir': 'uploads',
    'log_dir': 'logs',
    'cgi_dir': 'cgi-bin',
    'allow_dirlisting': True,
    'allow_cgi': True,
    'chunk_size': 8192,
    'index_files': ['index.html', 'index.htm'],
}

if os.path.exists('config.json'):
    with open('config.json', 'r') as fd:
        file_config = json.load(fd)
        Config.update(file_config)
else:
    with open('config.json', 'w') as fd:
        json.dump(Config, fd, 
                  indent=4, separators=(",", ": "), 
                  ensure_ascii=False, sort_keys=True)
